#ifndef __BUFFER_H
#define __BUFFER_H
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

/** @file
 * \brief A lightweight library for buffered input and output.
 *
 * libfwb buffer.h, FWB, http://www.tuxad.com
 *
 * Lightweight and fast alternative functions to
 * substitute widely used stdio-functions like
 * i.e. 'printf()'. The source code is bigger when
 * using libfwb-buffer but the compiled code misses
 * the bloated stdio code and gives you fewer syscalls.
 * Therefore the resulting code is smaller and runs faster.
 *
 * Example:
 * Assume you want to substitute this code line:
 * <pre>
 *   printf("%s = %u\n", a, b);
 * </pre>
 *
 * Then you can replace it with:
 * <pre>
 *   #include "lib/buffer.h"
 *   #define WRITEBUFSIZE 4096
 *   char buf1space[WRITEBUFSIZE+1];
 *   buffer buf1;
 *
 *   buffer_initwrite(&buf1, buf1space, WRITEBUFSIZE, 1);
 * 
 *   buffer_putstr(&buf1, a);
 *   buffer_putstr(&buf1, " = "); // or buffer_putmem(&buf1, " = ", 3);
 *   buffer_putuint32(&buf1, b);
 *   buffer_putc(&buf1, '\\n');
 *   buffer_flush(&buf1);
 * </pre>
 *
 * More example code:
 * 
 * <pre>
 *   #include "lib/buffer.h"
 *   #define STRINGSIZE 256
 *   #define READBUFSIZE 4096
 *   ...
 *   char buf0space[READBUFSIZE+1];
 *   buffer buf0;
 *   char line[STRINGSIZE];
 *
 *   buffer_initread(&buf0, buf0space, READBUFSIZE, 0);
 *   ...
 *   c = buffer_getline(&buf0, line, STRINGSIZE);
 *     if (c < 0) return 0; // error
 * </pre>
 *
 */

/** define an EOF if not defined */
#ifndef EOF
#define EOF (-1)
#endif


/** Short code for the struct.*/
typedef struct st_buffer buffer;

/** This is the structure used in this library.*/
struct st_buffer {
  /** first char in bufspace */
  char *s;
  /** cursor */
  char *c;
  /** first char after bufspace */
  char *t;
  /** the file descriptor */
  int fd;
  /** used just for read buffers: size of the buffer - it must be 0 for write buffers */
  uint32_t l;
};

/** initialize buffer to be used as output buffer */
static inline void buffer_initwrite(buffer *b, char *start_of_buffer, uint32_t len_of_buffer, int file_descriptor) {
  b->l = 0;
  if (len_of_buffer >= 1) {
    b->s = start_of_buffer;
    b->t = start_of_buffer + len_of_buffer;
    b->c = start_of_buffer;
    b->fd = file_descriptor;
  } else {
    b->s = b->t = b->c = 0;
    b->fd = -1;
  }
}
/** initialize buffer to be used as input buffer */
static inline void buffer_initread(buffer *b, char *start_of_buffer, uint32_t len_of_buffer, int file_descriptor) {
  if (len_of_buffer >= 1) {
    b->s = start_of_buffer;
    b->t = start_of_buffer + len_of_buffer;
    b->c = b->t;
    b->l = len_of_buffer;
    b->fd = file_descriptor;
  } else {
    b->s = b->t = b->c = 0;
    b->fd = -1;
  }
}
int buffer_flush(buffer *b);
int buffer_flushlines(buffer *b);
int buffer_putc(buffer *b, register const char a);
int buffer_putuint8(buffer *b, uint16_t num);
int buffer_putuint32(buffer *b, uint32_t num);
int buffer_putuint64(buffer *b, uint64_t num);
int buffer_getchar(buffer *b);
int buffer_getline(buffer *b, char *str, int str_size);

int buffer_putmem(buffer *b, register const char *a, register uint32_t l);
int buffer_putstr(buffer *b, register const char *a);

int buffer_openread(buffer *b, const char *filename);
int buffer_openwrite(buffer *b, const char *filename);
int buffer_close(buffer *b);
int buffer_catfromfile(buffer *b, const char *filename);
int buffer_filecmp(const char *filename1, const char *filename2);

#endif
