/**
 * threadcpu.c
 * Frank W. Bergmann, tuxad.com, 9/2018
 */
#include "config.h"
#include "../lib/libfwb.h"
#include "../lib/buffer.h"
#include <stdlib.h>
#include <stdint.h>
#include <sysexits.h>
#include <limits.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <dirent.h>
#include <sys/mman.h>

#define PIPEBUF_SIZE 4096
#define BUF_SIZE 4096
#define STRING_SIZE 255

#define STR1(s) #s
#define STR(s) STR1(s)

#ifdef CONFIG_THREADCPU_JSTACK_PATH
#define JSTACK_PATH CONFIG_THREADCPU_JSTACK_PATH
#else
#define JSTACK_PATH "/usr/bin/jstack"
#endif
#ifdef CONFIG_THREADCPU_JAVA_CHECK_FILE
#define JAVA_CHECK_FILE CONFIG_THREADCPU_JAVA_CHECK_FILE
#else
#define JAVA_CHECK_FILE "maps"
#endif
#ifdef CONFIG_THREADCPU_JAVA_CHECK_STRING
#define JAVA_CHECK_STRING CONFIG_THREADCPU_JAVA_CHECK_STRING
#else
#define JAVA_CHECK_STRING "/lib/rt.jar"
#endif

typedef struct st_thread_element {
  pid_t thread_id;
  char *thread_name;
  void *next_element;
} thread_element;

typedef struct st_task_element {
  pid_t task;
  long unsigned utime1;
  long unsigned stime1;
  long unsigned utime2;
  long unsigned stime2;
  void *next_element;
} task_element;

static const char usage[]=
  "\n"
  "threadcpu shows CPU usage of threads in user% and system%\n"
  "\n"
  "usage: threadcpu [-h] [-s seconds] [-p path-to-jstack]\n"
  "options:\n"
  "  -h display this help page\n"
  "  -s measuring interval in seconds, default: 10\n"
  "  -p path to JRE jstack, default: "JSTACK_PATH"\n"
  "example usage:\n"
  "  threadcpu -s 30 -p /opt/java/default/bin/jstack 2>/dev/null|sort -n|tail -n 12\n"
  "output columns:\n"
  "  user percent <SPACE> system percent <SPACE> PID/NID [ <SPACE> JVM thread name OR (process name) ]\n"
  "\n";

static const char jiffy_warning[] = 
  "WARNING: low jiffy count, you should consider increasing interval time\n";

static inline void printusage() {
  write1(usage, sizeof(usage)-1);
  _exit(EX_USAGE);
}

static inline void print(const char *str) {
  write1(str, strlen(str));
}

static pid_t pid_popen_child;
static int status_popen_child;

static int popen_argv(const char *argv[]) {
  int fd[2];

  if (pipe(fd)<0) return -3;
  if ((pid_popen_child=fork())<0) {
    close(fd[0]); close(fd[1]);
    return -4;
  }
  if (!pid_popen_child) {
    close(fd[0]); close(0);
    dup2(fd[1],1); close(fd[1]); 
    execve(argv[0],(char*const*)argv,NULL);
    close(1);
    return -2;
  }
  close(fd[1]);
  return fd[0];
}

static inline void popen_waitpid() {
  waitpid(pid_popen_child, &status_popen_child, 0);
}

static inline int hex2num(char hex) {
  int num = hex - '0';
  if (num > 9) num -= 7;
  if (num > 9) num -= 32;
  return num;
}

static pid_t nid2pid(const char *thread_str){
  pid_t pid = 4096 * hex2num(thread_str[0]);
  pid += 256 * hex2num(thread_str[1]);
  pid += 16 * hex2num(thread_str[2]);
  pid += hex2num(thread_str[3]);
  return pid;
}

static thread_element *add_thread(thread_element **list, thread_element *current_element, pid_t thread_id, const char *thread_name) {
  if (thread_name == 0) return current_element;
  if (thread_id == 0) return current_element;
  thread_element *new_element = alloc(sizeof(thread_element));
  if (new_element == 0) return current_element;
  char *new_thread_name = alloc(strlen(thread_name)+1);
  if (new_thread_name == 0) return current_element;
  if (__unlikely(*list == 0)) {
    /* empty list, current_element will also be 0 */
    *list = new_element;
  } else {
    current_element->next_element = new_element;
  }
  new_element->next_element = 0;
  strcpy(new_thread_name, thread_name);
  new_element->thread_name = new_thread_name;
  new_element->thread_id = thread_id;
  return new_element;
}

static thread_element *add_threads(const char *path_to_jstack, const char *pid_str, thread_element **list, thread_element *current_thread_element) {
  register char *p;
  register char *r;
  int c;
  char pipebuf[PIPEBUF_SIZE+1];
  char line[BUF_SIZE+1];
  int fd;
  static const char *jstack_argv[]={0,0,0};

  jstack_argv[0] = path_to_jstack;
  jstack_argv[1] = pid_str;
  fd = popen_argv(jstack_argv);
  if (fd < 0) goto out;

  p = pipebuf;
  buffer buf0 = { p, p + PIPEBUF_SIZE, p + PIPEBUF_SIZE, fd, PIPEBUF_SIZE };

  *line = 0;
  line[BUF_SIZE] = 0;
  for (;;) {
    c = buffer_getline(&buf0, line, BUF_SIZE);
    if (c == EOF) break;
    if ((p = strstr(line, " nid=0x")) != 0) {
      r = strchr(line + 1, '"');
      if (r == 0) continue;
      *r = 0;
      r = p + 7; p = r + 4; *p = 0;
      current_thread_element = add_thread(list, current_thread_element, nid2pid(r), line + 1);
      if (__unlikely(current_thread_element == 0)) break;
    }
  }

  close(fd);
  popen_waitpid();
  out:
  return current_thread_element;
}

static task_element *add_task(task_element **list, task_element *current_element, pid_t task) {
  if (task == 0) return current_element;
  task_element *new_element = alloc(sizeof(task_element));
  if (new_element == 0) return current_element;
  if (__unlikely(*list == 0)) {
    /* empty list, current_element will also be 0 */
    *list = new_element;
  } else {
    current_element->next_element = new_element;
  }
  new_element->next_element = 0;
  new_element->task = task;
  new_element->utime1 = 0;
  new_element->stime1 = 0;
  new_element->utime2 = 0;
  new_element->stime2 = 0;
  return new_element;
}

static task_element *find_task(task_element *list, pid_t task) {
  for (;;) {
    if (list == 0) return 0; if (list->task == task) return list; list = list->next_element;
    if (list == 0) return 0; if (list->task == task) return list; list = list->next_element;
    if (list == 0) return 0; if (list->task == task) return list; list = list->next_element;
    if (list == 0) return 0; if (list->task == task) return list; list = list->next_element;
  }
}

static const char *get_threadname(thread_element *list, pid_t thread_id) {
  thread_element *current_thread_element = list;
  for (;;) {
    if (current_thread_element == 0) return 0; if (current_thread_element->thread_id == thread_id) return current_thread_element->thread_name; current_thread_element = current_thread_element->next_element;
    if (current_thread_element == 0) return 0; if (current_thread_element->thread_id == thread_id) return current_thread_element->thread_name; current_thread_element = current_thread_element->next_element;
    if (current_thread_element == 0) return 0; if (current_thread_element->thread_id == thread_id) return current_thread_element->thread_name; current_thread_element = current_thread_element->next_element;
    if (current_thread_element == 0) return 0; if (current_thread_element->thread_id == thread_id) return current_thread_element->thread_name; current_thread_element = current_thread_element->next_element;
  }
}

static thread_element *add_pid_to_thread_list(thread_element **list, thread_element *current_element, const char *stat_file, const char *pid_str) {
  char buf[STRING_SIZE+1];
  uint64_t val64;
  char *p;
  char *t;
  str2uint64(pid_str, &val64);
  if (val64 == 0) return current_element;
  /* only add if it does not already exist, maybe it was already added as Java thread */
  if (get_threadname(*list, val64) != 0) return current_element;
  ssize_t len = file2buf(stat_file, buf, STRING_SIZE);
  if (len <= 0) return current_element;
  p = (char *) proc_stat_value(buf, PROC_PID_STAT_COMM);
  t = p;
  for (;;) {
    if (!*t) break;
    if (*t == ' ') { *t = 0; break; }
    ++t;
  }
  return add_thread(list, current_element, val64, p);
}

static uint64_t get_jiffies(const char *path, int field) {
  char buf[STRING_SIZE+1];
  uint64_t val64;
  ssize_t len;
  len = file2buf(path, buf, STRING_SIZE);
  if (len <= 0) return 0;
  str2uint64(proc_stat_value(buf, field), &val64);
  return val64;
}

static const char *strn0str(register const char* s, const char* t, size_t limit) {
  register char tc;
  register size_t tlen;

  if (limit == 0) return 0;
  tc = *t++;
  if (tc != 0) {
    tlen = strlen(t);
    for (;;) {
      for (;;) {
        if (limit-- < 1) return 0; if (*s++ == tc) break;
        if (limit-- < 1) return 0; if (*s++ == tc) break;
        if (limit-- < 1) return 0; if (*s++ == tc) break;
        if (limit-- < 1) return 0; if (*s++ == tc) break;
      }
      if (tlen > limit) return 0;
      if (strncmp(s, t, tlen) == 0) break;
    }
    --s;
  }
  return s;
}


int main(int argc,char *const argv[]) {
  register char *p;
  int c;
  char buf1space[BUF_SIZE+1];
  thread_element *thread_list = 0;
  thread_element *current_thread_element = 0;
  task_element *task_list = 0;
  task_element *current_task_element = 0;
  static const char jstack[]=JSTACK_PATH;
  const char *jstack_path;
  DIR *procdirp;
  struct dirent *procdirep;
  char procpath[PATH_MAX+1];
  char *procpathpid;
  char *procpathtask;
  DIR *taskdirp;
  struct dirent *taskdirep;
  unsigned long task;
  int sleep_seconds = 10;
  uint64_t user1;
  uint64_t system1;
  uint64_t user2;
  uint64_t system2;
  uint64_t user;
  uint64_t system;
  uint64_t val64;
  const char *rbuf;
  off_t rlen;

  p = buf1space; buffer buf1 = { p, p, p + BUF_SIZE, 1 };

  p = 0;
  jstack_path = jstack;
  for (c = 1; argc > c; ++c) {
    if (*argv[c] == '-') {
      if (argv[c][2] != '\0') printusage();
      switch (argv[c][1]) {
        case 'p': ++c;
                  jstack_path = argv[c];
                  break;
        case 's': ++c;
                  if (str2uint64(argv[c], &val64) != 0) sleep_seconds = val64;
                  break;
        default:  printusage();
      }
    } else {
      p = argv[c]; /* unused */
      break;
    }
  }

  /* step through proc dir first time */
  procpathpid = stpcpy(procpath, "/proc/");
  procdirp = opendir(procpath);
  if (procdirp == 0) _exit(EX_OSFILE);

  /* step through processes */
  for(;;) {

    procdirep = readdir(procdirp);
    if (procdirep == 0) break;
    /**
     * On opendir("/proc") the kernel won't list all possible entries.
     * But if you access directly a task dir it will show other virtual
     * dirs which are not shown on opendir()/readdir():
     * 
     * $ ls -U /proc/|grep 20413
     * $ ls -ld /proc/20413/stat
     * -r--r--r-- 1 root root 0 Sep 12 07:55 /proc/20413/stat
     * $ ls -U /proc/|grep 20413
     * $ ls -ld /proc/20413/stat
     * -r--r--r-- 1 root root 0 Sep 12 07:55 /proc/20413/stat
     * $ ls -U /proc/|grep 20413
     * $ ls -ld /proc/20413/stat
     * -r--r--r-- 1 root root 0 Sep 12 07:55 /proc/20413/stat
     * $ ls -U /proc/|grep 20413
     * $ ls -ld /proc/20413/stat
     * -r--r--r-- 1 root root 0 Sep 12 07:55 /proc/20413/stat
     */

    /* PID? */
    if ((*procdirep->d_name < '0') || (*procdirep->d_name > '9')) continue;

    p = stpcpy(procpathpid, procdirep->d_name);
    *p++ = '/'; *p = 0;

    /* java process detection */
    stpcpy(p, JAVA_CHECK_FILE);
    /* FIXME: lseek can't be used on proc fs */
    if ((char *) map_file(procpath, &rbuf, &rlen) != (char *) -1) {
      /* we need a faster / better way */
      if (strn0str(rbuf, JAVA_CHECK_STRING, rlen) != 0) {
        current_thread_element = add_threads(jstack_path, procdirep->d_name, &thread_list, current_thread_element);
        goto do_task_dir;
      }
      munmap((void *) rbuf, rlen);
    }
    stpcpy(p, "stat");
    current_thread_element = add_pid_to_thread_list(&thread_list, current_thread_element, procpath, procdirep->d_name);

    do_task_dir:
    procpathtask = stpcpy(p, "task/");

    /* step through tasks/threads */
    taskdirp = opendir(procpath);
    if (taskdirp == 0) goto next_process;

    for(;;) {

      taskdirep = readdir(taskdirp);
      if (taskdirep == 0) break;

      /* PID? */
      if ((*taskdirep->d_name < '0') || (*taskdirep->d_name > '9')) continue;

      str2uint64(taskdirep->d_name, &val64); task = val64;
      current_task_element = add_task(&task_list, current_task_element, task);
/*
      if (__unlikely(task_list == 0)) {
        task_list = current_task_element;
      }
*/
      p = stpcpy(procpathtask, taskdirep->d_name);
      p = stpcpy(p, "/stat");

      current_thread_element = add_pid_to_thread_list(&thread_list, current_thread_element, procpath, taskdirep->d_name);
      current_task_element->utime1 = get_jiffies(procpath, PROC_PID_STAT_UTIME);
      current_task_element->stime1 = get_jiffies(procpath, PROC_PID_STAT_STIME);

    }

    closedir(taskdirp);

    next_process:
    *procpathpid = 0;

  }

  closedir(procdirp);

  user1 = get_jiffies("/proc/stat",2);
  system1 = get_jiffies("/proc/stat",4);

/*
  // example debug output list of thread names:
  current_thread_element = thread_list;
  while (current_thread_element != 0) {
    printf("LISTPRINT %u %s\n", current_thread_element->thread_id, current_thread_element->thread_name);
    current_thread_element = current_thread_element->next_element;
  }
*/




  sleep(sleep_seconds);



  current_task_element = task_list;

  /* step through proc dir second time */
  procpathpid = stpcpy(procpath, "/proc/");
  procdirp = opendir(procpath);
  if (procdirp == 0) _exit(EX_OSFILE);

  /* step through processes */
  for(;;) {

    procdirep = readdir(procdirp);
    if (procdirep == 0) break;

    /* PID? */
    if ((*procdirep->d_name < '0') || (*procdirep->d_name > '9')) continue;

    p = stpcpy(procpathpid, procdirep->d_name);
    procpathtask = stpcpy(p, "/task/");

    /* step through tasks/threads */
    taskdirp = opendir(procpath);
    if (taskdirp == 0) goto next_process2;

    for(;;) {

      taskdirep = readdir(taskdirp);
      if (taskdirep == 0) break;

      /* PID? */
      if ((*taskdirep->d_name < '0') || (*taskdirep->d_name > '9')) continue;

      str2uint64(taskdirep->d_name, &val64); task = val64;
      current_task_element = find_task(task_list, task);
      p = stpcpy(procpathtask, taskdirep->d_name);
      p = stpcpy(p, "/stat");

      if (current_task_element != 0) {
        current_task_element->utime2 = get_jiffies(procpath, PROC_PID_STAT_UTIME);
        current_task_element->stime2 = get_jiffies(procpath, PROC_PID_STAT_STIME);
      }

    }

    closedir(taskdirp);

    next_process2:
    *procpathpid = 0;

  }

  closedir(procdirp);

  user2 = get_jiffies("/proc/stat",2);
  system2 = get_jiffies("/proc/stat",4);


  /* now all values are collected, do calculations and output */
  user = user2 - user1; if (user == 0) user = 1;
  system = system2 - system1; if (system == 0) system = 1;

  current_task_element = task_list;
  for (;;) {
    uint64_t percent;

    /* user% */
    if (current_task_element->utime1 > current_task_element->utime2) {
      percent = 0;
    } else {
      percent = (current_task_element->utime2 - current_task_element->utime1) * 100;
      percent /= user;
    }
    buffer_putuint32(&buf1, percent);
    buffer_putc(&buf1, ' ');

    /* system% */
    if (current_task_element->stime1 > current_task_element->stime2) {
      percent = 0;
    } else {
      percent = (current_task_element->stime2 - current_task_element->stime1) * 100;
      percent /= system;
    }
    buffer_putuint32(&buf1, percent);
    buffer_putc(&buf1, ' ');

    /* pid/nid/thread-id */
    buffer_putuint32(&buf1, current_task_element->task);

    /* java thread name available */
    p = (char *) get_threadname(thread_list, current_task_element->task);
    if (p != 0) {
      buffer_putc(&buf1, ' ');
      buffer_putstr(&buf1, p);
    }
    buffer_putc(&buf1, '\n');

    current_task_element = current_task_element->next_element;
    if (current_task_element == 0) break;
  }

  buffer_flush(&buf1);

  if ((system < 100) || (user < 100)) {
    write2(jiffy_warning, sizeof(jiffy_warning)-1);
  }

  _exit(0);

}
